/******************************************************************************
*
* project name: TI-MineSweeper
* filename:     timiner.h
* initial date: 04/07/2000
* authors:      thomas.nussbaumer@gmx.net (coding)
*               marcos.lopez@gmx.net (game play/graphics/BETA testing)
*               lionel_debroux@yahoo.fr (speedups/size optimizations)
*
* description:  header file inclued by C source files
*
* Current implementation uses TIGCC 0.95 Beta 15
*
* $Id: timiner.h,v 1.3 2004/07/14 15:08:58 DEBROUX Lionel Exp $
*
******************************************************************************/

/*---------------------------------------------------------------------------*/
/* some important defines                                                    */
/*---------------------------------------------------------------------------*/
#if !defined (__TIMINER_INCLUDE__)
#define __TIMINER_INCLUDE__

#include <tigcclib.h>
#include "extgraph.h"

#define PROGRAM_NAME     "TI-MineSweeper"
#define PROGRAM_VERSION  "1.32"
#define TITLE_AND_VERSION PROGRAM_NAME" v"PROGRAM_VERSION

#if ((defined(USE_TI89) && defined(USE_TI92PLUS)) || (defined(USE_TI89) && defined(USE_V200)))
#error cannot produce both versions at once
#elif defined(USE_TI89)
#define C89_92(x,y) (x)
#elif (defined(USE_TI92PLUS) || defined(USE_V200))
#define C89_92(x,y) (y)
#else
#error Please specify either -DUSE_TI89 or -DUSE_TI92P/-DUSE_V200 on commandline
#endif

#define PLANEOFFSET_X C89_92(0,40)
#define PLANEOFFSET_Y C89_92(0,14)

//-------------------
// special key values
//-------------------
#define KEY_HOME         277 // doesn't exist on TI92 !!

//-----------------------------
// some constants and variables
//-----------------------------
// counter.c
#define COUNTER_WIDTH    19
#define COUNTER_HEIGHT   15

#define NUMBER_HEIGHT  15
#define NUMBER_WIDTH    6
#define NUMBER_OFFSET  15

// tiles.c
#define TYPE_UNDISCOVERED   0
#define TYPE_MARKED         8
#define TYPE_INTERROGATION 16
#define TYPE_WRONGMARK     24
#define TYPE_MINE          32
#define TYPE_CLICKEDMINE   40

// smiley.c
#define SMILEY_WIDTH     17
#define SMILEY_HEIGHT    17

#define SMILEY_HAPPY     0
#define SMILEY_SURPRISED (1*17)
#define SMILEY_SAD       (2*17)
#define SMILEY_WIN       (3*17)

//----------------------
// maximum size of board
//----------------------
#define MAX_COLUMNS      18
#define MAX_LINES         8

#define nr_lines          8
#define board_height      (((nr_lines) << 3) + 34)
#define board_offset_y    ((LCD_HEIGHT - board_height) >> 1)
#define bc_ypos           board_offset_y + 7
#define tc_ypos           board_offset_y + 7
#define smiley_y          board_offset_y + 6
#define tiles_offset_y    board_offset_y + 29


/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/* internal used structures and related defines                              */
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
//-------------------------------
// structure describing a counter
//-------------------------------
typedef struct {
    short xpos;         // x position of counter
    //short ypos;         // y position of counter
    short value;        // active value of counter
    char digits[3];    // internally used
} counter_t;

//------------------------------
// description of one board tile
//------------------------------
typedef struct {
    char count;  // count of surrounding bombs (-1 == this tile is a bomb)
    char state;  // indicates state of the tile (see STATE_XXXX defines)
    char flag;   // flag used to mark tiles which graphics must be updated
    char unused;
} tile_t;

#define BOMB_VALUE       -1      // special count value which indicates a bomb

#define STATE_CLOSED      0      // tile is closed
#define STATE_OPENED      1      // tile is opened and it is a normal tile
#define STATE_MARKED      2      // tile is marked by user
#define STATE_INTERROGATE 3      // tile is "de-marked" by user
#define STATE_WRONGMARKED 4      // tile is wrongly marked
#define STATE_CLICKEDMINE 5      // tile is the bomb which was opened by the user

#define FLAG_NONE         0      // tile graphics is up-to-date
#define FLAG_TAGGED       0xFF   // tile graphics must be updated


register short cursor_x asm("%d6"); // x coordinate of cursor
register short cursor_y asm("%d7"); // y coordinate of cursor
register tile_t* board  asm("%a5"); // the board itself (allocated dynamically)

//---------------------------------------------
// describes one entry of the stack which is
// used by the "open board recursively" feature
//---------------------------------------------
typedef struct {
    short x;
    short y;
} stack_t;



// Macros and external functions, variables needed by them.

// counters.c
unsigned short firsttime = 1;

extern void UpdateCounter(counter_t *p);

/*===========================================================================*/
/* initialize counter (NOTE: counter structure must be setup already)        */
/*===========================================================================*/
#define InitCounter(p) UpdateCounter(p)



// cursor.c
extern SCR_RECT fullscreen;
extern short tiles_offset_x;

unsigned char pointer_plane1[] = {0x80,0xc0,0xe0,0xb0,0xb8,0x9c,0x9e,0xb8,0xdc,0x9c,0x0e,0x0e};
unsigned char pointer_plane2[] = {0x80,0xc0,0xa0,0x90,0x88,0x84,0x8e,0xa8,0xd4,0x94,0x0a,0x0e};
unsigned char mask_plane1[]    = {~0x00,~0x00,~0x40,~0x60,~0x70,~0x78,~0x70,~0x50,~0x08,~0x08,~0x04,~0x00};


SCR_RECT old_rect = {{255,255,255,255}};


extern unsigned char* cursor_buf1;
extern unsigned char* cursor_buf2;


/*===========================================================================*/
/* sets x0 to 255 to indicate that buffer contains no useful data            */
/*===========================================================================*/
#define InitCursorBuffer() { old_rect.xy.x0 = 255; }


/*===========================================================================*/
/* restores background of old cursor position                                */
/* restore only if buffer contains data (x0==255 indicates no data)          */
/*===========================================================================*/
#define RestoreOldCursorPos() ({ \
    if (old_rect.xy.x0 != 255) { \
        GraySetAMSPlane(0); \
        BitmapPut(old_rect.xy.x0,old_rect.xy.y0,cursor_buf1,&fullscreen,A_REPLACE); \
        GraySetAMSPlane(1); \
        BitmapPut(old_rect.xy.x0,old_rect.xy.y0,cursor_buf2,&fullscreen,A_REPLACE); \
    } \
})


/*===========================================================================*/
/* backups background of cursor location (given x/y is the logical position) */
/* First, calculate physical x/y position and setup rectangle to backup;     */
/* Second, get contents of both planes                                       */
/*===========================================================================*/
#define BackupCursorPos() ({ \
    short x = (cursor_x << 3) + tiles_offset_x + 3; \
    short y = (cursor_y << 3) + tiles_offset_y + 3; \
    old_rect.xy.x0 = x-1; \
    old_rect.xy.y0 = y-1; \
    old_rect.xy.x1 = x+9; \
    old_rect.xy.y1 = y+13; \
    GraySetAMSPlane(0); \
    BitmapGet(&old_rect,cursor_buf1); \
    GraySetAMSPlane(1); \
    BitmapGet(&old_rect,cursor_buf2); \
})


/*===========================================================================*/
/* output cursor sprite at logical position (x/y)                            */
/*===========================================================================*/
#define OutputCursor() ({\
    short x = (cursor_x << 3) + tiles_offset_x + 3; \
    short y = (cursor_y << 3) + tiles_offset_y + 3; \
    GraySprite8_MASK_R(x,y,12,pointer_plane1,pointer_plane2,mask_plane1,mask_plane1,GetPlane(LIGHT_PLANE),GetPlane(DARK_PLANE)); \
})


// tiles.c
extern short tiles_offset_x;

// using count_tiles2 for both planes
unsigned char count_tiles2[]={
0x00,0x01,0x00,0x01,0x00,0x01,0x00,0x55,  //no mines near
0x00,0x11,0x30,0x11,0x10,0x39,0x00,0x55,  //one mine near
0x00,0x31,0x08,0x11,0x20,0x39,0x00,0x55,  //two mines near
0x00,0x31,0x08,0x11,0x08,0x31,0x00,0x55,  //three mines near
0x00,0x21,0x28,0x39,0x08,0x09,0x00,0x55,  //four mines near
0x00,0x39,0x20,0x31,0x08,0x31,0x00,0x55,  //five mines near
0x00,0x19,0x20,0x39,0x28,0x39,0x00,0x55,  //six mines near
0x00,0x39,0x08,0x11,0x20,0x21,0x00,0x55,  //seven mines near
0x00,0x39,0x28,0x39,0x28,0x39,0x00,0x55}; //eight mines near

unsigned char misc_tiles1[] = {
0x01,0x7d,0x7d,0x7d,0x7d,0x7d,0x01,0xff,  //undiscovered
0x01,0x45,0x4d,0x7d,0x7d,0x7d,0x01,0xff,  //flag
0x01,0x7d,0x7d,0x7d,0x7d,0x7d,0x01,0xff,  //interrogation
0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,  //wrongly marked
0xff,0xff,0xff,0xef,0xff,0xff,0xff,0xff,  //mine
0x00,0x39,0x7c,0x6d,0x7c,0x39,0x00,0x55}; //clicked mine

unsigned char misc_tiles2[]={
0x01,0x03,0x03,0x03,0x03,0x03,0x7f,0xff,  //undiscovered
0x01,0x3b,0x3b,0x0b,0x0b,0x1f,0x7f,0xff,  //flag
0x01,0x33,0x0b,0x13,0x03,0x13,0x7f,0xff,  //interrogation
0x00,0x45,0x28,0x11,0x28,0x45,0x00,0x55,  //wrongly marked
0x00,0x39,0x7c,0x6d,0x7c,0x39,0x00,0x55,  //mine
0xff,0xff,0xff,0xef,0xff,0xff,0xff,0xff}; //clicked mine

unsigned char tiles_cleaner[]={
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};


/*===========================================================================*/
/* output single tile at position (x/y)                                      */
/*===========================================================================*/
void OutputTile(unsigned char* src1, unsigned char* src2, short x, short y) {
    x = x<<3;
    x += tiles_offset_x;
    y = y<<3;
    y += tiles_offset_y;

    GraySprite8_MASK_R(x,y,8,src1,src2,tiles_cleaner,tiles_cleaner,GetPlane(LIGHT_PLANE),GetPlane(DARK_PLANE));
}

/*===========================================================================*/
/* output a count tile for given count at position (x/y)                     */
/* NOTE: using count_tiles2 for both planes to get more contrast !!!         */
/*===========================================================================*/
#define OutputCountTile(count,x,y) ({ \
    OutputTile(count_tiles2+((count)<<3),count_tiles2+((count)<<3),x,y); \
})

/*===========================================================================*/
/* output a misc tile of given type at position (x/y)                        */
/*===========================================================================*/
void OutputMiscTile(short type,short x,short y) { 
    OutputTile(misc_tiles1+type,misc_tiles2+type,x,y);
}


// smiley.c
extern void OutputSmiley(short type);


// score.c
//---------------------------------------------
// length of playername array in highscore
// structure (don't change without increasing
// VERSION_FILEFORMAT)
//---------------------------------------------
#define PLAYER_NAME_LENGTH 20

//---------------------------------------------
// number of highscores
//---------------------------------------------
#define NR_HIGHSCORES 4

//---------------------------------------------
// structure to hold a highscore (don't change
// without increasing VERSION_FILEFORMAT)
//---------------------------------------------
typedef struct {
   short time;
   char  name[PLAYER_NAME_LENGTH];
} highscore_t;

extern highscore_t scores[NR_HIGHSCORES];

//---------------------------------------------
// if highscore_t is changed, you must change
// ALSO this define to make sure that highscore
// files which contains entries in the old
// format are not used !!!!
//---------------------------------------------
#define VERSION_FILEFORMAT 2


//---------------------------------------------
// default values of highscores for all modes
//---------------------------------------------
highscore_t defaults[NR_HIGHSCORES] = {
    {999,"       Programming "},
    {999,"  Thomas Nussbaumer"},
    {999,"   Gameplay and Gfx"},
    {999,"      Marcos Lopez "}
};


//---------------------------------------------
// active highscores are stored here !!
//---------------------------------------------
highscore_t scores[NR_HIGHSCORES] = {
    {999,"       Programming "},
    {999,"  Thomas Nussbaumer"},
    {999,"   Gameplay and Gfx"},
    {999,"      Marcos Lopez "}
};


//---------------------------------------------
// variable name to use for highscores
// a more or less confuse name is chosen to
// make relatively sure that it won't conflict
// with an existing variable on the calculator
//---------------------------------------------
#define HIGHSCORE_SYMNAME SYMSTR_CONST("timinhsc")

/*===========================================================================*/
/* returns true if highscore variable is archived                            */
/*===========================================================================*/
static inline short isArchivedOrLocked(void) {
    SYM_ENTRY* symentry;

    if (!(symentry = DerefSym(SymFind(HIGHSCORE_SYMNAME)))) return 0;

    return (symentry->flags.flags_n & (SF_ARCHIVED | SF_LOCKED));
}

/*===========================================================================*/
/* write highscores to variable (type == STR)                                */
/*===========================================================================*/
void WriteHighScores(void) {
    SYM_ENTRY* sym;
    short*      p;
    
    // 2 for size bytes,
    // 2 for header,
    // NR_HIGHSCORES*sizeof(highscore_t) for highscores,
    // 7 for end of file.
    // Rounded up to the next multiple of 2 by HeapAlloc.
#define size (2+2+NR_HIGHSCORES*(sizeof(highscore_t))+7)
    
    //------------------------------------------------
    // if highscore variable is archived or locked
    // -> do nothing
    //------------------------------------------------
    if (isArchivedOrLocked()) return;
    
    sym = DerefSym(SymAdd(HIGHSCORE_SYMNAME));
    
    if (sym) {
        sym->handle=HeapAlloc(size);
        if (sym->handle) {
            p = (short*)HeapDeref(sym->handle);
    
            *p++ = size-2;
            *p++ = VERSION_FILEFORMAT;
            memcpy(p,&scores,sizeof(highscore_t)*NR_HIGHSCORES);
            (unsigned char *)p += sizeof(highscore_t)*NR_HIGHSCORES;
            // Custom 'TICT' extension.
            *p++ = (0 << 8) + 'T';
            *p++ = ('I' << 8) + 'C';
            *p++ = ('T' << 8) + 0;
            *p++ = (OTH_TAG << 8) + 0;
        }
    }
#undef size
}


/*===========================================================================*/
/* load highscores from variable (type == STR)                               */
/*===========================================================================*/
static inline void LoadHighScores(void) {
    HSym           hsym;
    char          *p;

    //-------------------------------------------------------
    // if there isn't already a highscore file -> write a
    // default one
    //-------------------------------------------------------
    hsym=SymFind(HIGHSCORE_SYMNAME);
    if(!hsym.folder) {
        memcpy(scores,defaults,sizeof(highscore_t)*NR_HIGHSCORES);
        WriteHighScores();
        return;
    }
    
    //-------------------------------------------------------
    // If we come here we've found the highscore file
    //-------------------------------------------------------
    p = (char*)(HeapDeref(DerefSym(hsym)->handle));
    
    p += 3; // skip size and leading zero !

    //-------------------------------------------------------
    // check version number and load highscores if version
    // number is okay
    //-------------------------------------------------------
    if(*p == VERSION_FILEFORMAT) {
        p++;
        memcpy(&scores,p,sizeof(highscore_t)*NR_HIGHSCORES);
        return;
    }
    else {
        memcpy(scores,defaults,sizeof(highscore_t)*NR_HIGHSCORES);
        WriteHighScores();
    }
}

// logo.c
extern void Logo(void);


#endif
